import numpy as np
from numba import njit,jit,prange
import pandas as pd

@njit(parallel=False)
def _check_minp(win,minp,N):
    if minp > win:
        raise ValueError('minp must be <= win')
    elif minp > N:
        minp = N + 1
    elif minp == 0:
        minp = 1
    elif minp < 0:
        raise ValueError('minp must be >= 0')
    return minp

def roll_sum(input,win,minp):
    """
    Computes the rolling sum for a :code:`pandas` Series.

    Parameters
    ----------
    input: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).

    Returns
    -------
    output: pandas.core.frame.DataFrame
        A :code:`pandas` DataFrame with the rolling sum of :code:`input`.
    """
    _assertions(input,win,minp)
    return pd.DataFrame(_roll_sum(input.values,win,minp),input.index)

@njit(parallel=False)
def _roll_sum(input,win,minp):
    """
    Computes the rolling sum of a :code:`numpy` array.

    Parameters
    ----------
    input: np.array of type double
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).

    Returns
    -------
    output: np.array of type double
    """

    val = float(0) 
    prev = float(0)
    sum_x = float(0)
    nobs = int(0)
    N = int(len(input))

    minp = _check_minp(win, minp, N)   
    output = np.empty(N,dtype=np.float64)

    for i in range(minp - 1):
        val = input[i]

        if val == val:
            nobs += 1
            sum_x += val

        output[i] = np.nan
        

    for i in range(minp - 1,N):        
        val = input[i]

        if val == val:        
            nobs += 1
            sum_x += val

        if i > win - 1:
            prev = input[i - win]
            
            if prev == prev:
                sum_x -= prev                
                nobs -= 1

        if nobs >= minp:
            output[i] = sum_x
        else:
            output[i] = np.nan            

    return output            

def roll_mean(input,win,minp):
    """
    Computes the rolling mean of a :code:`pandas` Series.

    Parameters
    ----------
    input: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).

    Returns
    -------
    output: pandas.core.frame.DataFrame
        A :code:`pandas` DataFrame with the rolling mean of :code:`input`.
    """
    _assertions(input,win,minp)
    return pd.DataFrame(_roll_mean(input.values,win,minp),input.index)


@njit(parallel=False)
def _roll_mean(input,win,minp):
    """
    Computes the rolling mean of a numpy array.

    Parameters
    ----------
    input: np.array of type double
    win: int 
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).

    Returns
    -------
    output: np.array of type double
    """

    val = float(0)
    prev = float(0)
    sum_x = float(0)
    nobs = int(0)
    N = int(len(input))

    minp = _check_minp(win, minp, N)   
    output = np.empty(N,dtype=np.float64)

    for i in range(minp - 1):
        val = input[i]

        if val == val:
            nobs += 1
            sum_x += val

        output[i] = np.nan
        
    for i in range(minp - 1,N):        
        val = input[i]

        if val == val:        
            nobs += 1
            sum_x += val

        if i > win - 1:
            prev = input[i - win]
            
            if prev == prev:
                sum_x -= prev                
                nobs -= 1

        if nobs >= minp:
            output[i] = sum_x/nobs
        else:
            output[i] = np.nan            

    return output            


def roll_var(input,win,minp,ddof=1):
    """
    Computes the rolling variance for a :code:`pandas` Series.

    Parameters
    ----------
    input: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.

    Returns
    -------
    output: pandas.core.frame.DataFrame
        A :code:`pandas` DataFrame with the rolling variance of :code:`input`.
    """
    _assertions(input,win,minp,ddof=ddof)
    return pd.DataFrame(_roll_var(input.values,win,minp,ddof=ddof),input.index)


@njit(parallel=False)
def _roll_var(input,win,minp,ddof=1):
    """
    Computes the rolling variance of a numpy array.

    Parameters
    ----------
    input: np.array of type double
    win: int 
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NA).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.

    Returns
    -------
    output: np.array of type double
    """

    val = float(0)
    prev = float(0)
    sum_x = float(0)
    sum_xx = float(0)
    nobs = int(0)
    N = int(len(input))

    minp = _check_minp(win, minp, N)        
    output = np.empty(N,dtype=np.float64)

    for i in range(minp - 1):
        val = input[i]

        if val == val:
            nobs += 1
            sum_x += val
            sum_xx += val**2

        output[i] = np.nan


    for i in range(minp - 1,N):        
        val = input[i]

        if val == val:
            nobs += 1
            sum_x += val
            sum_xx += val**2

        if i > win - 1:
            prev = input[i - win]
            
            if prev == prev:
                sum_x -= prev
                sum_xx -= prev**2
                nobs -= 1

        if nobs >= minp:
            # pathological case
            if nobs == 1:
                output[i] = 0
                continue

            val = (nobs * sum_xx - sum_x**2) / (nobs * (nobs - ddof))
            if val < 0:
                val = 0

            output[i] = val
        else:
            output[i] = np.nan

    return output

def roll_std(input,win,minp,ddof=1):
    """
    Computes the rolling standard deviation for a :code:`pandas` Series.

    Parameters
    ----------
    input: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.

    Returns
    -------
    output: pandas.core.frame.DataFrame
        A :code:`pandas` DataFrame with the rolling standard deviation of :code:`input`.
    """
    _assertions(input,win,minp,ddof=ddof)
    return pd.DataFrame(_roll_std(input.values,win,minp,ddof=ddof),input.index)


@njit(parallel=False)
def _roll_std(input,win,minp,ddof=1):
    """
    Computes the rolling standard deviation of a numpy array.

    Parameters
    ----------
    input: np.array of type double
    win: int 
        Length of the moving window
    minp: int 
        Minimum number of observations in window required to have a value 
        (otherwise result is NA).
    ddof: int
        Delta cegrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.

    Returns
    -------
    output: np.array of type double
    """

    val = float(0)
    prev = float(0)
    sum_x = float(0)
    sum_xx = float(0)
    nobs = int(0)
    N = int(len(input))

    minp = _check_minp(win, minp, N)        
    output = np.empty(N,dtype=np.float64)

    for i in range(minp - 1):
        val = input[i]
        
        if val == val:
            nobs += 1
            sum_x += val
            sum_xx += val**2

        output[i] = np.nan

    for i in range(minp - 1,N):        
        val = input[i]

        if val == val:
            nobs += 1
            sum_x += val
            sum_xx += val**2

        if i > win - 1:
            prev = input[i - win]
            
            if prev == prev:
                sum_x -= prev
                sum_xx -= prev**2
                nobs -= 1

        if nobs >= minp:
            # pathological case
            if nobs == 1:
                output[i] = 0
                continue

            val = (nobs * sum_xx - sum_x**2) / (nobs * (nobs - ddof))
            if val < 0:
                val = 0

            output[i] = np.sqrt(val)
        else:
            output[i] = np.nan

    return output

def _assertions(input,win,minp,ddof=-999):
    assert (isinstance(input,pd.core.frame.DataFrame) or isinstance(input,pd.core.series.Series)), 'Input must be a pandas Series or DataFrame'
    assert ddof<input.shape[0], 'Invalid input for ddof: ddof must be less than the number of observations in input'
    if len(input.dtypes) > 0:
        assert all([pd.api.types.is_numeric_dtype(d) for d in input.dtypes]), 'All columns must have a numeric type.'
    else:
        assert pd.api.types.is_numeric_dtype(input.dtypes), 'All columns must be a numeric type.'

def _assertions2(x,y,win,minp,idx,ddof=-999):
    assert isinstance(x,pd.core.series.Series), 'x must be a pandas Series'
    assert isinstance(y,pd.core.series.Series), 'y must be a pandas Series'
    assert ddof<x.shape[0], 'Invalid input for ddof: ddof must be less than the number of observations in x and y'
    assert x.shape==y.shape, 'x and y must have the same shape'
    assert (idx=='x' or idx=='y'), 'Invalid input for idx'
    assert all([pd.api.types.is_numeric_dtype(d) for d in [x.dtype,y.dtype]]), 'All series must have a numeric type.'

def roll_cov(x,y,win,minp,ddof=1,idx='x'):
    """
    Computes the rolling covariance of two :code:`pandas` series.

    Parameters
    ----------
    x: pandas.core.series.Series
        This series must have strictly numeric type.
    y: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.
    idx: {'x','y'}
        Whether to use the index for x or for y for the return series. Defaults to 'x'.

    Returns
    -------
    output: pandas.core.series.Series
        A :code:`pandas` Series with the rolling covariance of x and y.
    """
    _assertions2(x,y,win,minp,idx,ddof=ddof)
    if idx=='x':
        myidx = x.index
    else: 
        myidx = y.index
    return pd.DataFrame(_roll_cov(x.values,y.values,win,minp,ddof=ddof),myidx)

@njit(parallel=False)
def _roll_cov(arr_x,arr_y,win,minp,ddof=1):
    """
    Computes the rolling covariance between two numpy arrays

    Parameters
    ----------
    arr_x: np.array of type double
    arr_y : np.array of type double
    win: int 
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NA).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.

    Returns
    -------
    output: np.array of type double
    """

    sum_x = float(0)
    sum_y = float(0)
    sum_xy = float(0)
    nobs = int(0)
    N = int(len(arr_x))

    minp = _check_minp(win, minp, N)        
    output = np.empty(N,dtype=np.float64)

    for i in range(minp - 1):    
        x = arr_x[i]
        y = arr_y[i]        

        if x == x and y == y:
            nobs += 1
            sum_x += x
            sum_y += y
            sum_xy += x*y

        output[i] = np.nan

    for i in range(minp - 1,N):        
        x = arr_x[i]
        y = arr_y[i]        

        if x == x and y == y:
            nobs += 1
            sum_x += x
            sum_y += y
            sum_xy += x*y

        if i > win - 1:
            prev_x = arr_x[i - win]
            prev_y = arr_y[i - win]
            
            if prev_x == prev_x and prev_y == prev_y:
                nobs -= 1
                sum_x -= prev_x
                sum_y -= prev_y                
                sum_xy -= prev_x*prev_y

        if nobs >= minp:
            # pathological case
            if nobs == 1:
                output[i] = 0
                continue

            output[i] = (nobs * sum_xy - sum_x*sum_y) / (nobs * (nobs - ddof))
        else:
            output[i] = np.nan

    return output

def roll_idio(y,x,win,minp,ddof=1,idx='x'):
    """
    Computes the rolling idio residual standard deviation from a univariate 
    regression, y = a + bx + e.

    Parameters
    ----------
    y: pandas.core.series.Series
        This series must have strictly numeric type.
    x: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.
    idx: {'x','y'}
        Whether to use the index for x or for y for the return series. Defaults to 'x'.

    Returns
    -------
    output: pandas.core.series.Series
        A :code:`pandas` Series with the rolling idio residual standard deviation from a univariate regression, y = a + bx + e.
    """
    _assertions2(x,y,win,minp,idx,ddof=ddof)
    if idx=='x':
        myidx = x.index
    else: 
        myidx = y.index
    return pd.DataFrame(_roll_idio(y.values,x.values,win,minp,ddof=ddof),myidx)


@njit(parallel=False)
def _roll_idio(arr_y,arr_x,win,minp,ddof=1):
    """
    Computes the rolling idio residual standard deviation from a univariate 
    regression, y = a + bx + e.

    Parameters
    ----------
    arr_y: np.array, dtype=double
        the dependent variable 
    arr_x: np.array, dtype=double
        the independent variable
    win: float
        Length of the moving window
    minp: float
        Minimum number of observations in window required to have a value 
        (otherwise result is NA).
    ddof: float 
        Delta cegrees of freedom. The divisor used in calculations is N - ddof,
        where N represents the number of elements.

    Returns
    -------
    out: np.array, dtype = double
    """
    
    var_y = _roll_var(arr_y,win,minp,ddof)
    var_x = _roll_var(arr_x,win,minp,ddof)
    cov   = _roll_cov(arr_y,arr_x,win,minp,ddof)    

    N = int(len(arr_x))
    out = np.empty(N,dtype=np.float64)

    for i in range(0,N):
        out[i] = np.sqrt(var_y[i] - cov[i]**2/var_x[i])

    return out

def roll_beta(y,x,win,minp,ddof=1,idx='x'):
    """
    Computes the rolling estimated slope coefficient (beta)  from an univariate regression, y = a + bx + e. 

    Parameters
    ----------
    y: pandas.core.series.Series
        This series must have strictly numeric type.
    x: pandas.core.series.Series
        This series must have strictly numeric type.
    win: int
        Length of the moving window
    minp: int
        Minimum number of observations in window required to have a value 
        (otherwise result is NaN).
    ddof: int
        Delta degrees of freedom. The divisor used in calculations is N - ddof, 
        where N represents the number of elements.
    idx: {'x','y'}
        Whether to use the index for x or for y for the return series. Defaults to 'x'.

    Returns
    -------
    output: pandas.core.series.Series
        A :code:`pandas` Series with the rolling estimated slope coefficient (beta)  from an univariate regression, y = a + bx + e. 
    """
    _assertions2(x,y,win,minp,idx,ddof=ddof)
    if idx=='x':
        myidx = x.index
    else: 
        myidx = y.index
    return pd.DataFrame(_roll_beta(y.values,x.values,win,minp,ddof=ddof),myidx)

@njit(parallel=False)
def _roll_beta(arr_y,arr_x,win,minp,ddof=1):
    """
    Computes the rolling estimated slope coefficient (beta)  from an 
    univariate regression, y = a + bx + e. 

    Parameters
    ----------
    arr_y: np.array, dtype=double
        the dependent variable 
    arr_x: np.array, dtype=double
        the independent variable
    win: float
        Length of the moving window
    minp: float
        Minimum number of observations in window required to have a value 
        (otherwise result is NA).
    ddof: float 
        Delta cegrees of freedom. The divisor used in calculations is N - ddof,
        where N represents the number of elements.

    Returns
    -------
    out: np.array, dtype = double
    """
    
    var_x = _roll_var(arr_x,win,minp,ddof)
    cov   = _roll_cov(arr_y,arr_x,win,minp,ddof)    

    N = int(len(arr_x))
    out = np.empty(N,dtype=np.float64)

    for i in range(N):
        out[i] = cov[i]/var_x[i]

    return out
    
import multiprocessing
import joblib as jl
from numba import njit

@njit(parallel=False)
def _ols_numba2(d,yvari,xvari):
    #assert np.linalg.matrix_rank(d[:,xvari])==xvari.shape[0], 'Independent variable matrix is not full column rank. Check data carefully.'
    gamma,nul1,nul2,nul3 = np.linalg.lstsq(d[:,xvari],d[:,yvari],rcond=-1)
    return gamma
    
@njit(parallel=False)
def _residvol_numba(d,yvari,xvari,gamma):
    r = np.empty(d.shape[0])
    for i in range(d.shape[0]):
        r[i] = d[i,yvari]-d[i,xvari].dot(gamma)
    return np.array([r.std(),np.NaN])

def _numba_assertions(data):
    if len(data.dtypes) > 0:
        assert all([pd.api.types.is_numeric_dtype(d) for d in data.dtypes]), 'All columns must have a numeric type.'
    else:
        assert pd.api.types.is_numeric_dtype(data.dtypes), 'All columns must have a numeric type.'
        
def _assertions(data,yvar,xvar,intercept,backend,roll,residuals):
    assert isinstance(data,pd.core.frame.DataFrame), 'Invalid input for `data`.'
    assert isinstance(xvar,list), 'Invalid input for `xvar`.'
    assert isinstance(intercept,bool), 'Invalid input for `intercept`.'
    assert hasattr(yvar,'__iter__')==False or isinstance(yvar,str), 'Invalid input for `yvar`. `yvar` must either be a string or not be iterable'
    assert backend in ['loky','multiprocessing','threading'], 'Invalid input for `backend`.'
    assert isinstance(roll,int), 'Invalid input for `roll`. `roll` must be an integer.'
    assert (residuals=='residvol') or isinstance(residuals,bool), "Invalid input for `residuals`. `residuals` must be boolean or 'residvol'." 


def rolling_multiple(data,yvar,xvar,roll,intercept=True,residuals=False,backend='loky'):
    """
    Rolling multiple regression implementation using :code:`pandas` groupby for grouping, linear algebra routines
    compiled with :code:`numba` for regressions, and :code:`joblib` for parallelization. Jobs are pre-dispatched 
    to each core for performance.
    
    Parameters
    ----------
    
    data: pandas.core.frame.DataFrame
        A dataframe with regressand and regressors for multiple regression. This dataframe must have strictly numeric 
        types.
    yvar: str
        The name of the regressand variable in :code:`data`. 
    xvar: list(str)
        A list of the names of the regressor variables in :code:`data`. 
    roll: int
        The number of observations (rows in data) over which to roll, inclusive of the current row. 
        For example, if :code:`roll=120`, the rolling regression will include the current observation and the previous
        119 observations.
    intercept: bool
        Whether or not to regress with an intercept.
    residuals: {bool,'residvol'}
        Whether or not to include residuals in the output dataframe (boolean). If :code:`'residvol'` is input, the standard
        deviation of the residuals from the rolling regression will be output. If :code:`True` is input, the residual corresponding
        to the observation will be output.
    backend: {'loky','multiprocessing','threading'}
        The joblib backend to use for parallel processing. 'loky' is used by default and is recommended.
        
    Returns 
    -------
    
    A pandas DataFrame which contains rolling regression coefficients and residuals, if specified.
    
    """
        
    _assertions(data,yvar,xvar,intercept,backend,roll,residuals)
    _numba_assertions(data)
    
    if intercept==True:
        data['intercept'] = 1
        xvar.insert(0,'intercept')
    
    yvari = data.columns.get_loc(yvar)
    xvari = [data.columns.get_loc(x) for x in xvar]
    roller = roll-1
    grouped = [np.array(data.loc[j-roller:j,:]) for j in range(roller,data.shape[0])] 
    n_jobs = multiprocessing.cpu_count()
    batch_size = int(len(grouped) / n_jobs) 
    retlst = jl.Parallel(n_jobs=n_jobs,backend=backend,batch_size=batch_size,pre_dispatch='all') \
                (jl.delayed(_ols_numba2)(g,int(yvari),np.array(xvari)) for g in grouped)
    d = pd.DataFrame(np.vstack(retlst))
    
    if residuals=='residvol':
        #retlst2 = jl.Parallel(n_jobs=n_jobs,backend=backend,batch_size=batch_size,pre_dispatch='all') \
        #         (jl.delayed(_residvol_numba)(g,int(yvari),np.array(xvari),np.array(d.iloc[idx,:])) for idx,g in enumerate(grouped))
        retlst2 = [_residvol_numba(g,int(yvari),np.array(xvari),np.array(d.iloc[idx,:])) for idx,g in enumerate(grouped)]
        rvol = pd.DataFrame(np.vstack(retlst2)[:,0])
    
    fill = pd.DataFrame(np.NaN,index=[i for i in range(roller)],columns=xvar)
    d.columns = xvar 
    d = pd.concat([fill,d],axis=0).reset_index(drop=True)
    
    if residuals != False:
        
        if residuals==True:
            if intercept:
                data['yhat'] = data['intercept']
            else:
                data['yhat'] = 0
            for x in xvar:
                data['yhat'] = data['yhat'] + data[x]*d[x]
            d['resid'] = data[yvar] - data['yhat']
        
        elif residuals=='residvol':
            fill = pd.DataFrame(np.NaN,index=[i for i in range(roller)],columns=[0])
            rvol = pd.concat([fill,rvol],axis=0).reset_index(drop=True)
            d['residvol'] = rvol
            
    return d

 
